import functools
import json
import logging
import re
from io import BytesIO
from urllib.parse import parse_qs, urlparse

from pytest import fixture, mark, raises
from tornado.httpclient import HTTPResponse
from tornado.httputil import HTTPHeaders
from traitlets.config import Config

from ..gitlab import GitLabOAuthenticator
from .mocks import setup_oauth_mock

API_ENDPOINT = f"/api/v{GitLabOAuthenticator().gitlab_api_version}"


id_to_username_map = {}


def user_model(username):
    """Return a user model"""

    # generate an id based on the username hash and remember it
    id = abs(hash(username)) % (10**8)
    id_to_username_map[id] = username
    return {
        "username": username,
        "id": id,
    }


@fixture
def gitlab_client(client):
    setup_oauth_mock(
        client,
        host='gitlab.com',
        access_token_path='/oauth/token',
        user_path=API_ENDPOINT + '/user',
    )
    return client


def mock_api_version(client, version):
    def mock_version_response(request):
        ret = {'version': version, 'revision': "f79c1794977"}
        return HTTPResponse(
            request,
            200,
            headers={'Content-Type': 'application/json'},
            buffer=BytesIO(json.dumps(ret).encode('utf-8')),
        )

    regex = re.compile(API_ENDPOINT + '/version')
    client.hosts['gitlab.com'].append((regex, mock_version_response))


@mark.parametrize(
    "test_variation_id,class_config,expect_allowed,expect_admin",
    [
        # no allow config tested
        ("00", {}, False, None),
        # allow config, individually tested
        ("01", {"allow_all": True}, True, None),
        ("02", {"allowed_users": {"user1"}}, True, None),
        ("03", {"allowed_users": {"not-test-user"}}, False, None),
        ("04", {"admin_users": {"user1"}}, True, True),
        ("05", {"admin_users": {"not-test-user"}}, False, None),
        # allow config, some combinations of two tested
        (
            "10",
            {
                "allow_all": False,
                "allowed_users": {"not-test-user"},
            },
            False,
            None,
        ),
        (
            "11",
            {
                "admin_users": {"user1"},
                "allowed_users": {"not-test-user"},
            },
            True,
            True,
        ),
    ],
)
async def test_gitlab(
    gitlab_client,
    test_variation_id,
    class_config,
    expect_allowed,
    expect_admin,
):
    print(f"Running test variation id {test_variation_id}")
    c = Config()
    c.GitLabOAuthenticator = Config(class_config)
    c.GitLabOAuthenticator.username_claim = "username"
    authenticator = GitLabOAuthenticator(config=c)

    handled_user_model = user_model("user1")
    handler = gitlab_client.handler_for_user(handled_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)

    if expect_allowed:
        assert auth_model
        assert set(auth_model) == {"name", "admin", "auth_state"}
        assert auth_model["admin"] == expect_admin
        auth_state = auth_model["auth_state"]
        assert json.dumps(auth_state)
        assert "access_token" in auth_state
        user_info = auth_state[authenticator.user_auth_state_key]
        assert user_info == handled_user_model
        assert auth_model["name"] == user_info[authenticator.username_claim]
    else:
        assert auth_model == None


def make_link_header(urlinfo, page):
    return {
        "Link": f'<{urlinfo.scheme}://{urlinfo.netloc}{urlinfo.path}?page={page}>;rel="next"'
    }


@mark.parametrize(
    "paginate",
    [
        False,
        True,
    ],
)
async def test_allowed_groups(gitlab_client, paginate):
    authenticator = GitLabOAuthenticator()
    mock_api_version(gitlab_client, '12.4.0-ee')

    ## set up fake Gitlab API

    user_groups = {
        "user1": ["group0", "group1"],
    }

    groups_members_api_regex = re.compile(
        API_ENDPOINT + r'/groups/(.*)/members/all/(.*)'
    )

    def mocked_groups_members_api(request):
        """
        Is user_id a member of a group?

        https://docs.gitlab.com/ee/api/members.html#get-a-member-of-a-group-or-project
        is mocked solely by the HTTP response code.
        """
        urlinfo = urlparse(request.url)
        group, user_id = groups_members_api_regex.match(urlinfo.path).group(1, 2)
        username = id_to_username_map[int(user_id)]
        if group in user_groups.get(username, []):
            return HTTPResponse(request, 200)
        else:
            return HTTPResponse(request, 404)

    def mocked_groups_api(paginate, request):
        """
        What groups are the user that makes a request to the /groups API part
        of?

        https://docs.gitlab.com/ee/api/groups.html#list-groups is mocked by
        returning a list of dictionaries like {"path": <group_name>}, and only
        one group per page.
        """
        urlinfo = urlparse(request.url)
        _, token = request._headers.get('Authorization').split()
        username = gitlab_client.access_tokens[token]['username']
        if not paginate:
            return [{'path': group} for group in user_groups[username]]
        else:
            page = parse_qs(urlinfo.query).get('page', ['1'])
            page = int(page[0])
            return _mocked_groups_api_paginated(
                username, page, urlinfo, functools.partial(HTTPResponse, request)
            )

    def _mocked_groups_api_paginated(username, page, urlinfo, response):
        """
        Helper function for mocked_groups_api.
        """
        if page < len(user_groups[username]):
            headers = make_link_header(urlinfo, page + 1)
        elif page == len(user_groups[username]):
            headers = {}
        else:
            return response(400)

        headers.update({'Content-Type': 'application/json'})
        ret = [{'path': user_groups[username][page - 1]}]
        return response(
            200,
            headers=HTTPHeaders(headers),
            buffer=BytesIO(json.dumps(ret).encode('utf-8')),
        )

    gitlab_client.hosts['gitlab.com'].append(
        (groups_members_api_regex, mocked_groups_members_api)
    )
    gitlab_client.hosts['gitlab.com'].append(
        (API_ENDPOINT + '/groups', functools.partial(mocked_groups_api, paginate))
    )

    ## actual tests

    authenticator.allowed_gitlab_groups = ["group1"]

    handled_user_model = user_model("user1")
    handler = gitlab_client.handler_for_user(handled_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)
    assert auth_model
    assert json.dumps(auth_model["auth_state"])

    handled_user_model = user_model("user-not-in-group")
    handler = gitlab_client.handler_for_user(handled_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)
    assert auth_model is None


async def test_allowed_project_ids(gitlab_client):
    authenticator = GitLabOAuthenticator()
    mock_api_version(gitlab_client, '12.4.0-pre')

    non_project_member_user_model = user_model('non-project-member')
    guest_user_model = user_model('guest')
    developer_user_model = user_model('developer')
    user_projects = {
        '1': {
            str(guest_user_model["id"]): {
                'id': guest_user_model["id"],
                'name': guest_user_model["username"],
                'username': guest_user_model["username"],
                'state': 'active',
                'avatar_url': 'https://secure.gravatar.com/avatar/382a6b306679b2d97b547bfff3d73242?s=80&d=identicon',
                'web_url': f'https://gitlab.com/{guest_user_model["username"]}',
                'access_level': 10,  # Guest
                'expires_at': '2040-02-23',
            },
            str(developer_user_model["id"]): {
                'id': developer_user_model["id"],
                'name': developer_user_model["username"],
                'username': developer_user_model["username"],
                'state': 'active',
                'avatar_url': 'https://secure.gravatar.com/avatar/382a6b306679b2d97b547bfff3d73242?s=80&d=identicon',
                'web_url': f'https://gitlab.com/{guest_user_model["username"]}',
                'access_level': 30,  # Developer
                'expires_at': '2040-02-23',
            },
        }
    }

    projects_members_api_regex = re.compile(
        API_ENDPOINT + r'/projects/(.*)/members/all/(.*)'
    )

    def mocked_projects_members_api(request):
        """
        Is user_id a member of a project?

        https://docs.gitlab.com/ee/api/members.html#get-a-member-of-a-group-or-project
        is mocked by somewhat realistic response.
        """
        urlinfo = urlparse(request.url)
        project_id, user_id = projects_members_api_regex.match(urlinfo.path).group(1, 2)

        if user_projects.get(project_id) and user_projects[project_id].get(user_id):
            res = user_projects[project_id][user_id]
            return HTTPResponse(
                request=request,
                code=200,
                buffer=BytesIO(json.dumps(res).encode('utf8')),
                headers={'Content-Type': 'application/json'},
            )
        else:
            return HTTPResponse(request=request, code=404, buffer=BytesIO(b''))

    gitlab_client.hosts['gitlab.com'].append(
        (projects_members_api_regex, mocked_projects_members_api)
    )

    authenticator.allowed_project_ids = [1]

    # Forbidden, user doesn't have access to a project in allowed_project_ids
    handler = gitlab_client.handler_for_user(non_project_member_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)
    assert auth_model is None

    # Forbidden, user only has has guest access a project in allowed_project_ids
    handler = gitlab_client.handler_for_user(guest_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)
    assert auth_model is None

    # Authorized, user has developer access a project in allowed_project_ids
    handler = gitlab_client.handler_for_user(developer_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)
    assert auth_model
    assert json.dumps(auth_model["auth_state"])

    # Forbidden, project doesn't exist
    authenticator.allowed_project_ids = [0]
    handler = gitlab_client.handler_for_user(developer_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)
    assert auth_model is None

    # Authorized, user has developer access to one of the allowed_project_ids
    authenticator.allowed_project_ids = [0, 1]
    handler = gitlab_client.handler_for_user(developer_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)
    assert auth_model
    assert json.dumps(auth_model["auth_state"])


@mark.parametrize(
    "test_variation_id,class_config,expect_config,expect_loglevel,expect_message",
    [
        (
            "gitlab_group_whitelist",
            {"gitlab_group_whitelist": {"dummy"}},
            {"allowed_gitlab_groups": {"dummy"}},
            logging.WARNING,
            "GitLabOAuthenticator.gitlab_group_whitelist is deprecated in GitLabOAuthenticator 0.12.0, use GitLabOAuthenticator.allowed_gitlab_groups instead",
        ),
    ],
)
async def test_deprecated_config(
    caplog,
    test_variation_id,
    class_config,
    expect_config,
    expect_loglevel,
    expect_message,
):
    """
    Tests that a warning is emitted when using a deprecated config and that
    configuring the old config ends up configuring the new config.
    """
    print(f"Running test variation id {test_variation_id}")
    c = Config()
    c.GitLabOAuthenticator = Config(class_config)

    test_logger = logging.getLogger('testlog')
    if expect_loglevel == logging.ERROR:
        with raises(ValueError, match=expect_message):
            GitLabOAuthenticator(config=c, log=test_logger)
    else:
        authenticator = GitLabOAuthenticator(config=c, log=test_logger)
        for key, value in expect_config.items():
            assert getattr(authenticator, key) == value

    captured_log_tuples = caplog.record_tuples
    print(captured_log_tuples)

    expected_log_tuple = (test_logger.name, expect_loglevel, expect_message)
    assert expected_log_tuple in captured_log_tuples



import json
from io import BytesIO
from unittest.mock import Mock
from urllib.parse import parse_qs

from pytest import fixture, mark, raises
from tornado import web
from tornado.httpclient import HTTPResponse
from traitlets.config import Config

from ..globus import GlobusLogoutHandler, GlobusOAuthenticator
from ..oauth2 import STATE_COOKIE_NAME
from .mocks import mock_handler, setup_oauth_mock


def user_model(username, **kwargs):
    """Return a user model"""
    return {
        "preferred_username": username,
        **kwargs,
    }


def revoke_token_request_handler(request):
    assert request.method == 'POST', request.method
    auth_header = request.headers.get('Authorization')
    if auth_header:
        resp = BytesIO(json.dumps({'active': False}).encode('utf8'))
        return HTTPResponse(request=request, code=200, buffer=resp)
    else:
        return HTTPResponse(request=request, code=401)


@fixture
def mock_globus_token_response():
    return {
        'access_token': 'de48bedc44b79937f7aa67',
        'id_token': 'ClRha2UgbXkgbG92ZSwgdGFrZSBteSBsYW5kClRha2UgbWUgd2hlcmUgSS'
        'BjYW5ub3Qgc3RhbmQKSSBkb24ndCBjYXJlIGNhdXNlIEknbSBzdGlsbCBm'
        'cmVlCllvdSBjYW4ndCB0YWtlIHRoZSBza3kgZnJvbSBtZQpUYWtlIG1lIG'
        '91dCwgdG8gdGhlIGJsYWNrClRlbGwgJ2VtIEkgYWluJ3QgY29taW5nIGJh'
        'Y2sKQnVybiB0aGUgbGFuZCBhbmQgYm9pbCB0aGUgc2VhCllvdSBjYW4ndC'
        'B0YWtlIHRoZSBza3kgZnJvbSBtZQpUaGVyZSdzIG5vIHBsYWNlIEkgY2Fu'
        'IGJlClNpbmNlIEkgZm91bmQgc2VyZW5pdHkKWW91IGNhbid0IHRha2UgdG'
        'hlIHNreSBmcm9tIG1lCg==',
        'expires_in': 172800,
        'resource_server': 'auth.globus.org',
        'token_type': 'Bearer',
        'state': '5a5929fa3c0210042c2fbb455e1e39d0',
        'other_tokens': [
            {
                'access_token': 'fceb9836f9b6d1ae7d',
                'expires_in': 172800,
                'resource_server': 'transfer.api.globus.org',
                'token_type': 'Bearer',
                'state': '5a5929fa3c0210042c2fbb455e1e39d0',
                'scope': 'urn:globus:auth:scope:transfer.api.globus.org:all',
            },
            {
                'access_token': '309f9e6367d1ffffae0da625cb87b9ac543ee72a',
                'expires_in': 172800,
                'resource_server': 'groups.api.globus.org',
                'token_type': 'Bearer',
                'scope': 'urn:globus:auth:scope:groups.api.globus.org:view_my_groups_and_memberships',
            },
        ],
        'scope': 'profile openid',
    }


def get_groups_request_handler(request):
    mock_globus_groups_response = [
        {
            # group's IDs should really be UUIDs, but a simpler string is used
            # for consistency between tests
            'id': 'group1',
            'my_memberships': [{'role': 'member'}],
        },
    ]
    assert request.method == 'GET', request.method
    resp = BytesIO(json.dumps(mock_globus_groups_response).encode('utf-8'))
    return HTTPResponse(
        request=request,
        code=200,
        headers={'Content-Type': 'application/json'},
        buffer=resp,
    )


@fixture
def globus_tokens_by_resource_server(mock_globus_token_response):
    token_attrs = [
        'expires_in',
        'resource_server',
        'scope',
        'token_type',
        'refresh_token',
        'access_token',
    ]
    auth_token_dict = {
        attr_name: mock_globus_token_response.get(attr_name)
        for attr_name in token_attrs
    }
    other_tokens = [
        {attr_name: token_dict.get(attr_name) for attr_name in token_attrs}
        for token_dict in mock_globus_token_response['other_tokens']
    ]
    tokens = other_tokens + [auth_token_dict]
    return {token_dict['resource_server']: token_dict for token_dict in tokens}


def set_extended_token_response(client, host, access_token_path, new_token_response):
    """The default client fixture does a nice job of checking the access code
    response while returning tokens in the oauth2 spec, but Globus returns
    a bunch of other tokens, including an id_token. We want to make sure we
    capture the full Globus token response. This will attach the dict
    new_token_response to the built-in test response if it returns successfully"""
    # Find the existing endpoint, function pair in client.hosts
    url, func = next(
        filter(lambda host: host[0] == access_token_path, client.hosts[host])
    )

    # Wrap the built-in token response with our custom response, but only if
    # it returns successfully with an access token!
    def custom_token_response(request):
        response = func(request)
        if response.get('access_token'):
            # The original access_token is checked,
            new_token_response['access_token'] = response['access_token']
            return new_token_response
        else:
            return response

    # Return all existing paths with the addition of our custom wrapped handler.
    hosts = filter(lambda chost: chost[0] != access_token_path, client.hosts[host])
    client.add_host(host, [(url, custom_token_response)] + list(hosts))
    return client


@fixture
def globus_client(client, mock_globus_token_response):
    setup_oauth_mock(
        client,
        host='auth.globus.org',
        access_token_path='/v2/oauth2/token',
        user_path='/v2/oauth2/userinfo',
        token_type='bearer',
    )
    set_extended_token_response(
        client, 'auth.globus.org', '/v2/oauth2/token', mock_globus_token_response
    )
    client.add_host(
        'groups.api.globus.org',
        [
            ('/v2/groups/my_groups', get_groups_request_handler),
        ],
    )
    return client


@fixture
def mock_globus_user(globus_tokens_by_resource_server):
    class User:
        name = 'Wash'
        state = {'tokens': globus_tokens_by_resource_server}

        async def get_auth_state(self):
            return self.state

        async def save_auth_state(self, state):
            self.state = state

    return User()


@mark.parametrize(
    "test_variation_id,class_config,expect_allowed,expect_admin",
    [
        # no allow config tested
        ("00", {}, False, None),
        # allow config, individually tested
        ("01", {"allow_all": True}, True, None),
        ("02", {"allowed_users": {"user1"}}, True, None),
        ("03", {"allowed_users": {"not-test-user"}}, False, None),
        ("04", {"admin_users": {"user1"}}, True, True),
        ("05", {"admin_users": {"not-test-user"}}, False, None),
        ("06", {"allowed_globus_groups": {"group1"}}, True, None),
        ("07", {"allowed_globus_groups": {"test-user-not-in-group"}}, False, None),
        ("08", {"admin_globus_groups": {"group1"}}, True, True),
        ("09", {"admin_globus_groups": {"test-user-not-in-group"}}, False, False),
        # allow config, some combinations of two tested
        (
            "10",
            {
                "allow_all": False,
                "allowed_users": {"not-test-user"},
            },
            False,
            None,
        ),
        (
            "11",
            {
                "allowed_users": {"not-test-user"},
                "admin_users": {"user1"},
            },
            True,
            True,
        ),
        (
            "12",
            {
                "allowed_globus_groups": {"group1"},
                "admin_globus_groups": {"group1"},
            },
            True,
            True,
        ),
        (
            "13",
            {
                "allowed_globus_groups": {"group1"},
                "admin_globus_groups": {"test-user-not-in-group"},
            },
            True,
            False,
        ),
        (
            "14",
            {
                "allowed_globus_groups": {"test-user-not-in-group"},
                "admin_globus_groups": {"group1"},
            },
            True,
            True,
        ),
        (
            "15",
            {
                "allowed_globus_groups": {"test-user-not-in-group"},
                "admin_globus_groups": {"test-user-not-in-group"},
            },
            False,
            False,
        ),
        (
            "16",
            {
                "admin_users": {"user1"},
                "admin_globus_groups": {"group1"},
            },
            True,
            True,
        ),
        (
            "17",
            {
                "admin_users": {"user1"},
                "admin_globus_groups": {"test-user-not-in-group"},
            },
            True,
            True,
        ),
        (
            "18",
            {
                "admin_users": {"not-test-user"},
                "admin_globus_groups": {"group1"},
            },
            True,
            True,
        ),
        (
            "19",
            {
                "admin_users": {"not-test-user"},
                "admin_globus_groups": {"test-user-not-in-group"},
            },
            False,
            False,
        ),
    ],
)
async def test_globus(
    globus_client,
    test_variation_id,
    class_config,
    expect_allowed,
    expect_admin,
):
    print(f"Running test variation id {test_variation_id}")
    c = Config()
    c.GlobusOAuthenticator = Config(class_config)
    c.GlobusOAuthenticator.username_claim = "preferred_username"
    authenticator = GlobusOAuthenticator(config=c)

    handled_user_model = user_model("user1")
    handler = globus_client.handler_for_user(handled_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)

    if expect_allowed:
        assert auth_model
        assert set(auth_model) == {"name", "admin", "auth_state"}
        assert auth_model["admin"] == expect_admin
        auth_state = auth_model["auth_state"]
        assert json.dumps(auth_state)
        assert "tokens" in auth_state
        assert "transfer.api.globus.org" in auth_state["tokens"]
        user_info = auth_state[authenticator.user_auth_state_key]
        assert auth_model["name"] == user_info[authenticator.username_claim]
        if authenticator.allowed_globus_groups or authenticator.admin_globus_groups:
            assert auth_state["globus_groups"] == ["group1"]
    else:
        assert auth_model == None


@mark.parametrize(
    "name, allowed",
    [
        ("allowed", True),
        ("notallowed", False),
    ],
)
async def test_check_allowed_no_auth_state(name, allowed):
    authenticator = GlobusOAuthenticator(allowed_users={"allowed"})
    # allow check always gets called with no auth model during Hub startup
    # these are previously-allowed users who should pass until subsequent
    # this check is removed in JupyterHub 5
    assert await authenticator.check_allowed(name, None)


async def test_globus_pre_spawn_start(mock_globus_user):
    authenticator = GlobusOAuthenticator()
    spawner = Mock()
    spawner.environment = {}
    await authenticator.pre_spawn_start(mock_globus_user, spawner)
    assert 'GLOBUS_DATA' in spawner.environment


async def test_globus_defaults():
    authenticator = GlobusOAuthenticator()
    print(f"userdata_url: {authenticator.userdata_url}")
    print(f"authorize_url: {authenticator.authorize_url}")
    print(f"revocation_url: {authenticator.revocation_url}")
    print(f"token_url: {authenticator.token_url}")

    assert all(
        'https://auth.globus.org' in url
        for url in [
            authenticator.userdata_url,
            authenticator.authorize_url,
            authenticator.revocation_url,
            authenticator.token_url,
        ]
    )
    assert authenticator.scope == [
        'openid',
        'profile',
        'urn:globus:auth:scope:transfer.api.globus.org:all',
    ]


async def test_restricted_domain(globus_client):
    c = Config()
    c.GlobusOAuthenticator.allow_all = True
    c.GlobusOAuthenticator.identity_provider = "allowed.example.com"
    authenticator = GlobusOAuthenticator(config=c)

    handled_user_model = user_model("user1@example.com")
    handler = globus_client.handler_for_user(handled_user_model)
    with raises(web.HTTPError) as exc:
        await authenticator.get_authenticated_user(handler, None)
    assert exc.value.status_code == 403


async def test_namespaced_domain(globus_client):
    c = Config()
    c.GlobusOAuthenticator.allow_all = True
    authenticator = GlobusOAuthenticator(config=c)

    handled_user_model = user_model('wash@legitshipping.com@serenity.com')
    handler = globus_client.handler_for_user(handled_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)
    assert auth_model['name'] == 'wash'


async def test_username_from_email_scope_added(globus_client):
    c = Config()
    c.GlobusOAuthenticator.username_from_email = True
    authenticator = GlobusOAuthenticator(config=c)
    assert authenticator.scope == [
        'openid',
        'profile',
        'urn:globus:auth:scope:transfer.api.globus.org:all',
        'email',
    ]


async def test_username_from_email_restricted_pass(globus_client):
    c = Config()
    c.GlobusOAuthenticator.allow_all = True
    c.GlobusOAuthenticator.username_from_email = True
    c.GlobusOAuthenticator.identity_provider = "allowed.example.com"
    authenticator = GlobusOAuthenticator(config=c)

    handled_user_model = user_model(
        'dummy@example.com', email='user1@allowed.example.com'
    )
    handler = globus_client.handler_for_user(handled_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)
    assert auth_model
    assert auth_model["name"] == "user1"


async def test_username_from_email_restricted_fail(globus_client):
    c = Config()
    c.GlobusOAuthenticator.allow_all = True
    c.GlobusOAuthenticator.username_from_email = True
    c.GlobusOAuthenticator.identity_provider = "allowed.example.com"
    authenticator = GlobusOAuthenticator(config=c)

    handled_user_model = user_model(
        "user1@allowed.example.com", email="dummy@example.com"
    )
    handler = globus_client.handler_for_user(handled_user_model)
    with raises(web.HTTPError) as exc:
        await authenticator.get_authenticated_user(handler, None)
    assert exc.value.status_code == 403


async def test_token_exclusion(globus_client):
    c = Config()
    c.GlobusOAuthenticator.allow_all = True
    c.GlobusOAuthenticator.exclude_tokens = [
        "auth.globus.org",
        "groups.api.globus.org",
        "transfer.api.globus.org",
    ]
    authenticator = GlobusOAuthenticator(config=c)

    handled_user_model = user_model("user1@example.com")
    handler = globus_client.handler_for_user(handled_user_model)
    auth_model = await authenticator.get_authenticated_user(handler, None)
    assert auth_model
    assert auth_model['auth_state']
    assert not auth_model['auth_state']['tokens']


async def test_revoke_tokens(globus_client, mock_globus_user):
    # Wrap the revocation host to 'revoke' tokens by setting them in user auth
    # state. This way, we can get feedback to tell if the token was actually
    # sent to our 'host'
    def tok_revoke(request):
        resp = revoke_token_request_handler(request)
        token = parse_qs(request.body.decode('utf8'))['token'][0]
        for token_dict in mock_globus_user.state['tokens'].values():
            if token_dict['access_token'] == token:
                token_dict['access_token'] = 'token_revoked'
            if token_dict['refresh_token'] == token:
                token_dict['refresh_token'] = 'token_revoked'
        return resp

    # Add the token revocation endpoint. It's the only revocation endpoint we need.
    globus_client.add_host('auth.globus.org', [('/v2/oauth2/token/revoke', tok_revoke)])
    # Add refresh tokens to ensure those get revoked too.
    mock_globus_user.state['tokens']['auth.globus.org'][
        'refresh_token'
    ] = 'my_active_auth_refresh_token'
    mock_globus_user.state['tokens']['transfer.api.globus.org'][
        'refresh_token'
    ] = 'my_active_transfer_refresh_token'

    # Revoke the tokens!
    authenticator = GlobusOAuthenticator()
    await authenticator.revoke_service_tokens(mock_globus_user.state['tokens'])

    # Check tokens were properly revoked.
    user_tokens = mock_globus_user.state['tokens']
    assert user_tokens['auth.globus.org']['access_token'] == 'token_revoked'
    assert user_tokens['auth.globus.org']['access_token'] == 'token_revoked'
    assert user_tokens['transfer.api.globus.org']['access_token'] == 'token_revoked'
    assert user_tokens['transfer.api.globus.org']['access_token'] == 'token_revoked'


async def test_custom_logout(monkeypatch, mock_globus_user):
    custom_logout_url = 'https://universityofindependence.edu/logout'
    authenticator = GlobusOAuthenticator()
    logout_handler = mock_handler(GlobusLogoutHandler, authenticator=authenticator)
    monkeypatch.setattr(web.RequestHandler, 'redirect', Mock())
    logout_handler.clear_login_cookie = Mock()
    logout_handler.clear_cookie = Mock()
    logout_handler.get_current_user = Mock(return_value=mock_globus_user)
    logout_handler._jupyterhub_user = mock_globus_user
    monkeypatch.setitem(logout_handler.settings, 'statsd', Mock())

    # Sanity check: Ensure the logout handler and url are set on the hub
    handlers = [handler for _, handler in authenticator.get_handlers(None)]
    assert any([h == GlobusLogoutHandler for h in handlers])
    assert authenticator.logout_url('http://myhost') == 'http://myhost/logout'

    # Test the logout handler uses the custom URL
    authenticator.logout_redirect_url = custom_logout_url
    await logout_handler.get()
    logout_handler.redirect.assert_called_once_with(custom_logout_url)
    assert logout_handler.clear_login_cookie.called
    logout_handler.clear_cookie.assert_called_once_with(STATE_COOKIE_NAME)


async def test_logout_revokes_tokens(globus_client, monkeypatch, mock_globus_user):
    globus_client.add_host(
        'auth.globus.org', [('/v2/oauth2/token/revoke', revoke_token_request_handler)]
    )
    authenticator = GlobusOAuthenticator()
    logout_handler = mock_handler(GlobusLogoutHandler, authenticator=authenticator)

    # Setup
    monkeypatch.setattr(web.RequestHandler, 'redirect', Mock())
    logout_handler.get_current_user = Mock(return_value=mock_globus_user)
    logout_handler._jupyterhub_user = mock_globus_user
    monkeypatch.setitem(logout_handler.settings, 'statsd', Mock())
    monkeypatch.setitem(logout_handler.settings, 'login_url', '')

    logout_handler.clear_login_cookie = Mock()
    authenticator.revoke_tokens_on_logout = True

    await logout_handler.get()
    auth_state = await mock_globus_user.get_auth_state()
    assert auth_state == {'tokens': {}}


async def test_group_scope_added(globus_client):
    authenticator = GlobusOAuthenticator()
    authenticator.allowed_globus_groups = {'group-manager'}
    assert authenticator.scope == [
        'openid',
        'profile',
        'urn:globus:auth:scope:transfer.api.globus.org:all',
        'urn:globus:auth:scope:groups.api.globus.org:view_my_groups_and_memberships',
    ]



"""
A JupyterHub authenticator class for use with Bitbucket as an identity provider.
"""
import os

from jupyterhub.auth import LocalAuthenticator
from tornado.httputil import url_concat
from traitlets import Set, default

from .oauth2 import OAuthenticator


class BitbucketOAuthenticator(OAuthenticator):
    client_id_env = "BITBUCKET_CLIENT_ID"
    client_secret_env = "BITBUCKET_CLIENT_SECRET"
    user_auth_state_key = "bitbucket_user"

    @default("login_service")
    def _login_service_default(self):
        return os.environ.get("LOGIN_SERVICE", "Bitbucket")

    @default("authorize_url")
    def _authorize_url_default(self):
        return "https://bitbucket.org/site/oauth2/authorize"

    @default("token_url")
    def _token_url_default(self):
        return "https://bitbucket.org/site/oauth2/access_token"

    @default("userdata_url")
    def _userdata_url_default(self):
        return "https://api.bitbucket.org/2.0/user"

    allowed_teams = Set(
        config=True,
        help="""
        Allow members of selected Bitbucket teams to sign in.
        """,
    )

    # _deprecated_oauth_aliases is used by deprecation logic in OAuthenticator
    _deprecated_oauth_aliases = {
        "team_whitelist": ("allowed_teams", "0.12.0"),
        **OAuthenticator._deprecated_oauth_aliases,
    }
    team_whitelist = Set(
        config=True,
        help="""
        .. deprecated:: 0.12

           Use :attr:`allowed_teams`.
        """,
    )

    async def _fetch_user_teams(self, access_token, token_type):
        """
        Get user's team memberships via bitbucket's API.
        """
        headers = self.build_userdata_request_headers(access_token, token_type)
        next_page = url_concat(
            "https://api.bitbucket.org/2.0/workspaces", {'role': 'member'}
        )

        user_teams = set()
        while next_page:
            resp_json = await self.httpfetch(next_page, method="GET", headers=headers)
            next_page = resp_json.get('next', None)
            user_teams |= {entry["name"] for entry in resp_json["values"]}
        return user_teams

    async def update_auth_model(self, auth_model):
        """
        Fetch and store `user_teams` in auth state if `allowed_teams` is
        configured.
        """
        user_teams = set()
        if self.allowed_teams:
            access_token = auth_model["auth_state"]["token_response"]["access_token"]
            token_type = auth_model["auth_state"]["token_response"]["token_type"]
            user_teams = await self._fetch_user_teams(access_token, token_type)
        # sets are not JSONable, cast to list for auth_state
        auth_model["auth_state"]["user_teams"] = list(user_teams)

        return auth_model

    async def check_allowed(self, username, auth_model):
        """
        Overrides the OAuthenticator.check_allowed to also allow users part of
        `allowed_teams`.
        """
        if await super().check_allowed(username, auth_model):
            return True

        if self.allowed_teams:
            user_teams = set(auth_model["auth_state"].get("user_teams", []))
            if user_teams & self.allowed_teams:
                return True

        # users should be explicitly allowed via config, otherwise they aren't
        return False


class LocalBitbucketOAuthenticator(LocalAuthenticator, BitbucketOAuthenticator):
    """A version that mixes in local system user creation"""



import json
import re
import time
from unittest.mock import Mock

import jwt
import requests_mock
from pytest import fixture, mark
from tornado import web
from traitlets.config import Config

from ..mediawiki import AUTH_REQUEST_COOKIE_NAME, MWOAuthenticator
from .mocks import mock_handler


@fixture
def mediawiki():
    def post_token(request, context):
        authorization_header = request.headers['Authorization'].decode('utf8')
        request_nonce = re.search(r'oauth_nonce="(.*?)"', authorization_header).group(1)
        content = jwt.encode(
            {
                'username': 'user1',
                'aud': 'client_id',
                'iss': 'https://meta.wikimedia.org',
                'iat': time.time(),
                'nonce': request_nonce,
            },
            'client_secret',
        ).encode()

        return content

    with requests_mock.Mocker() as mock:
        mock.post(
            '/w/index.php?title=Special%3AOAuth%2Finitiate',
            text='oauth_token=key&oauth_token_secret=secret',
        )
        mock.post(
            '/w/index.php?title=Special%3AOAuth%2Ftoken',
            text='oauth_token=key&oauth_token_secret=secret',
        )
        mock.post('/w/index.php?title=Special%3AOAuth%2Fidentify', content=post_token)
        yield mock


@mark.parametrize(
    "test_variation_id,class_config,expect_allowed,expect_admin",
    [
        # no allow config tested
        ("00", {}, False, None),
        # allow config, individually tested
        ("01", {"allow_all": True}, True, None),
        ("02", {"allowed_users": {"user1"}}, True, None),
        ("03", {"allowed_users": {"not-test-user"}}, False, None),
        ("04", {"admin_users": {"user1"}}, True, True),
        ("05", {"admin_users": {"not-test-user"}}, False, None),
        # allow config, some combinations of two tested
        (
            "10",
            {
                "allow_all": False,
                "allowed_users": {"not-test-user"},
            },
            False,
            None,
        ),
        (
            "11",
            {
                "admin_users": {"user1"},
                "allowed_users": {"not-test-user"},
            },
            True,
            True,
        ),
    ],
)
async def test_mediawiki(
    mediawiki,
    test_variation_id,
    class_config,
    expect_allowed,
    expect_admin,
):
    print(f"Running test variation id {test_variation_id}")
    c = Config()
    c.MWOAuthenticator = Config(class_config)
    c.MWOAuthenticator.client_id = "client_id"
    c.MWOAuthenticator.client_secret = "client_secret"
    c.MWOAuthenticator.username_claim = "username"
    authenticator = MWOAuthenticator(config=c)

    handler = Mock(
        spec=web.RequestHandler,
        get_secure_cookie=Mock(return_value=json.dumps(['key', 'secret'])),
        request=Mock(query='oauth_token=key&oauth_verifier=me'),
        find_user=Mock(return_value=None),
    )
    auth_model = await authenticator.get_authenticated_user(handler, None)

    if expect_allowed:
        assert auth_model
        assert set(auth_model) == {"name", "admin", "auth_state"}
        assert auth_model["admin"] == expect_admin
        auth_state = auth_model["auth_state"]
        assert json.dumps(auth_state)
        assert "ACCESS_TOKEN_KEY" in auth_state
        assert "ACCESS_TOKEN_SECRET" in auth_state
        user_info = auth_state[authenticator.user_auth_state_key]
        assert auth_model["name"] == user_info[authenticator.username_claim]
    else:
        assert auth_model == None


async def test_login_redirect(mediawiki):
    authenticator = MWOAuthenticator(
        client_id='client_id',
        client_secret='client_secret',
    )
    record = []
    handler = mock_handler(
        authenticator.login_handler,
        'https://hub.example.com/hub/login',
        authenticator=authenticator,
    )
    handler.write = lambda buf: record.append(buf)
    await handler.get()
    assert handler.get_status() == 302
    assert 'Location' in handler._headers
    assert handler._headers['Location'].startswith(authenticator.mw_index_url)
    assert 'Set-Cookie' in handler._headers
    assert AUTH_REQUEST_COOKIE_NAME in handler._headers['Set-Cookie']



"""
A JupyterHub authenticator class for use with Google as an identity provider.
"""
import os

from jupyterhub.auth import LocalAuthenticator
from tornado.auth import GoogleOAuth2Mixin
from tornado.web import HTTPError
from traitlets import Dict, List, Set, Unicode, default, validate

from .oauth2 import OAuthenticator


class GoogleOAuthenticator(OAuthenticator, GoogleOAuth2Mixin):
    user_auth_state_key = "google_user"

    @default("login_service")
    def _login_service_default(self):
        return os.environ.get("LOGIN_SERVICE", "Google")

    @default("scope")
    def _scope_default(self):
        return ["openid", "email"]

    @default("username_claim")
    def _username_claim_default(self):
        return "email"

    @default("authorize_url")
    def _authorize_url_default(self):
        return "https://accounts.google.com/o/oauth2/v2/auth"

    google_api_url = Unicode(
        config=True,
        help="""
        Used to determine the default values for `token_url` and `userdata_url`.
        """,
    )

    @default("google_api_url")
    def _google_api_url(self):
        """get default google apis url from env"""
        google_api_url = os.getenv('GOOGLE_API_URL')

        # default to googleapis.com
        if not google_api_url:
            google_api_url = 'https://www.googleapis.com'

        return google_api_url

    @default("token_url")
    def _token_url_default(self):
        return f"{self.google_api_url}/oauth2/v4/token"

    @default("userdata_url")
    def _userdata_url_default(self):
        return f"{self.google_api_url}/oauth2/v1/userinfo"

    google_service_account_keys = Dict(
        Unicode(),
        config=True,
        help="""
        Service account keys to use with each domain, see https://developers.google.com/admin-sdk/directory/v1/guides/delegation

        Required if and only if `allowed_google_groups` or `admin_google_groups`
        is configured.
        """,
    )

    gsuite_administrator = Dict(
        Unicode(),
        config=True,
        help="""
        Username of a G Suite Administrator for the service account to act as.

        Required if and only if `allowed_google_groups` or `admin_google_groups`
        is configured.
        """,
    )

    allowed_google_groups = Dict(
        Set(Unicode()),
        config=True,
        help="""
        Allow members of selected Google groups to sign in.

        Use of this requires configuration of `gsuite_administrator` and
        `google_service_account_keys`.
        """,
    )

    admin_google_groups = Dict(
        Set(Unicode()),
        config=True,
        help="""
        Allow members of selected Google groups to sign in and consider them as
        JupyterHub admins.

        If this is set and a user isn't part of one of these groups or listed in
        `admin_users`, a user signing in will have their admin status revoked.

        Use of this requires configuration of `gsuite_administrator` and
        `google_service_account_keys`.
        """,
    )

    hosted_domain = List(
        Unicode(),
        config=True,
        help="""
        This config has two functions.

        1. Restrict sign-in to a list of email domain names, such as
           `["mycollege.edu"]` or `["college1.edu", "college2.edu"]`.
        2. If a single domain is specified, the username will be stripped to exclude the `@domain` part.

        Note that users with email domains in this list must still be allowed
        via another config, such as `allow_all`, `allowed_users`, or
        `allowed_google_groups`.

        ```{warning} Disruptive config changes
        Changing this config either to or from having a single entry is a
        disruptive change as the same Google user will get a new username,
        either without or with a domain name included.
        ```
        """,
    )

    @default('hosted_domain')
    def _hosted_domain_from_env(self):
        domains = []
        for domain in os.environ.get('HOSTED_DOMAIN', '').lower().split(';'):
            if domain:
                # check falsy to avoid trailing separators
                # adding empty domains
                domains.append(domain)
        return domains

    @validate('hosted_domain')
    def _cast_hosted_domain(self, proposal):
        """handle backward-compatibility with hosted_domain is a single domain as a string"""
        if isinstance(proposal.value, str):
            # pre-0.9 hosted_domain was a string
            # set it to a single item list
            # (or if it's empty, an empty list)
            if proposal.value == '':
                return []
            return [proposal.value.lower()]
        return [hd.lower() for hd in proposal.value]

    # _deprecated_oauth_aliases is used by deprecation logic in OAuthenticator
    _deprecated_oauth_aliases = {
        "google_group_whitelist": ("allowed_google_groups", "0.12.0"),
        **OAuthenticator._deprecated_oauth_aliases,
    }
    google_group_whitelist = Dict(
        config=True,
        help="""
        .. deprecated:: 0.12

           Use :attr:`allowed_google_groups`.
        """,
    )

    def user_info_to_username(self, user_info):
        """
        Overrides the default implementation to conditionally also strip the
        user email's domain name from the username based on the hosted_domain
        configuration. The domain saved to user_info for use by authorization
        logic.
        """
        username = super().user_info_to_username(user_info)
        user_email = user_info["email"]
        user_domain = user_info["domain"] = user_email.split("@")[1].lower()

        if len(self.hosted_domain) == 1 and self.hosted_domain[0] == user_domain:
            # unambiguous domain, use only base name
            username = username.split("@")[0]

        return username

    async def update_auth_model(self, auth_model):
        """
        Fetch and store `google_groups` in auth state if `allowed_google_groups`
        or `admin_google_groups` is configured.

        Sets admin status to True or False if `admin_google_groups` is
        configured and the user isn't part of `admin_users`. Note that leaving
        it at None makes users able to retain an admin status while setting it
        to False makes it be revoked.

        Strips the domain from the username if `hosted_domain` is configured
        with a single entry.
        """
        user_info = auth_model["auth_state"][self.user_auth_state_key]
        user_email = user_info["email"]
        user_domain = user_info["domain"]

        user_groups = set()
        if self.allowed_google_groups or self.admin_google_groups:
            user_groups = self._fetch_user_groups(user_email, user_domain)
        # sets are not JSONable, cast to list for auth_state
        user_info["google_groups"] = list(user_groups)

        if auth_model["admin"]:
            # auth_model["admin"] being True means the user was in admin_users
            return auth_model

        if self.admin_google_groups:
            # admin status should in this case be True or False, not None
            admin_groups = self.admin_google_groups.get(user_domain, set())
            auth_model["admin"] = bool(user_groups & admin_groups)

        return auth_model

    async def check_allowed(self, username, auth_model):
        """
        Overrides the OAuthenticator.check_allowed to also allow users part of
        `allowed_google_groups`.
        """
        # A workaround for JupyterHub < 5.0 described in
        # https://github.com/jupyterhub/oauthenticator/issues/621
        if auth_model is None:
            return True

        # before considering allowing a username by being recognized in a list
        # of usernames or similar, we must ensure that the authenticated user
        # has a verified email and is part of hosted_domain if configured.
        user_info = auth_model["auth_state"][self.user_auth_state_key]
        user_email = user_info["email"]
        user_domain = user_info["domain"]

        if not user_info["verified_email"]:
            message = f"Login with unverified email {user_email} is not allowed"
            self.log.warning(message)
            raise HTTPError(403, message)

        # NOTE: If hosted_domain is configured as ["a.com", "b.com"], and
        #       allowed_google_groups is declared as {"a.com": {"a-group"}}, a
        #       "b.com" user won't be authorized unless allowed in another way.
        #
        #       This means that its not possible to allow all users of a given
        #       domain if one wants to restrict another.
        #
        if self.hosted_domain:
            if user_domain not in self.hosted_domain:
                message = f"Login with domain @{user_domain} is not allowed"
                self.log.warning(message)
                raise HTTPError(403, message)

        if await super().check_allowed(username, auth_model):
            return True

        if self.allowed_google_groups:
            user_groups = set(user_info["google_groups"])
            allowed_groups = self.allowed_google_groups.get(user_domain, set())
            if user_groups & allowed_groups:
                return True

        # users should be explicitly allowed via config, otherwise they aren't
        return False

    def _service_client_credentials(self, scopes, user_email_domain):
        """
        Return a configured service client credentials for the API.
        """
        try:
            from google.oauth2 import service_account
        except:
            raise ImportError(
                "Could not import google.oauth2's service_account,"
                "you may need to run 'pip install oauthenticator[googlegroups]' or not declare google groups"
            )

        gsuite_administrator_email = "{}@{}".format(
            self.gsuite_administrator[user_email_domain], user_email_domain
        )
        self.log.debug(f"scopes are {scopes}, user_email_domain is {user_email_domain}")
        credentials = service_account.Credentials.from_service_account_file(
            self.google_service_account_keys[user_email_domain], scopes=scopes
        )

        credentials = credentials.with_subject(gsuite_administrator_email)

        return credentials

    def _service_client(self, service_name, service_version, credentials, http=None):
        """
        Return a configured service client for the API.
        """
        try:
            from googleapiclient.discovery import build
        except:
            raise ImportError(
                "Could not import googleapiclient.discovery's build,"
                "you may need to run 'pip install oauthenticator[googlegroups]' or not declare google groups"
            )

        self.log.debug(
            f"service_name is {service_name}, service_version is {service_version}"
        )

        return build(
            serviceName=service_name,
            version=service_version,
            credentials=credentials,
            cache_discovery=False,
            http=http,
        )

    def _fetch_user_groups(self, user_email, user_email_domain, http=None):
        """
        Return a set with the google groups a given user is a member of
        """
        # FIXME: When this function is used and waiting for web request
        #        responses, JupyterHub gets blocked from doing other things.
        #        Ideally the web requests should be made using an async client
        #        that can be awaited while JupyterHub handles other things.
        #
        credentials = self._service_client_credentials(
            scopes=[f"{self.google_api_url}/auth/admin.directory.group.readonly"],
            user_email_domain=user_email_domain,
        )
        service = self._service_client(
            service_name='admin',
            service_version='directory_v1',
            credentials=credentials,
            http=http,
        )

        resp = service.groups().list(userKey=user_email).execute()
        user_groups = {
            g['email'].split('@')[0] for g in resp.get('groups', [{'email': None}])
        }
        self.log.debug(f"user_email {user_email} is a member of {user_groups}")
        return user_groups


class LocalGoogleOAuthenticator(LocalAuthenticator, GoogleOAuthenticator):
    """A version that mixes in local system user creation"""



"""
A JupyterHub authenticator class for use with MediaWiki as an identity provider.
"""
import json
import os
from asyncio import wrap_future
from concurrent.futures import ThreadPoolExecutor

from jupyterhub.handlers import BaseHandler
from jupyterhub.utils import url_path_join
from mwoauth import ConsumerToken, Handshaker
from mwoauth.tokens import RequestToken
from traitlets import Any, Integer, Unicode, default

from oauthenticator import OAuthCallbackHandler, OAuthenticator

# Name of cookie used to pass auth token between the oauth
# login and authentication phase
AUTH_REQUEST_COOKIE_NAME = 'mw_oauth_request_token_v2'


# Helpers to jsonify/de-jsonify request_token
# It is a named tuple with bytestrings, json.dumps balks
def jsonify(request_token):
    return json.dumps(
        [
            request_token.key,
            request_token.secret,
        ]
    )


def dejsonify(js):
    key, secret = json.loads(js)
    return RequestToken(key, secret)


class MWLoginHandler(BaseHandler):
    async def get(self):
        consumer_token = ConsumerToken(
            self.authenticator.client_id,
            self.authenticator.client_secret,
        )

        handshaker = Handshaker(self.authenticator.mw_index_url, consumer_token)

        redirect, request_token = await wrap_future(
            self.authenticator.executor.submit(handshaker.initiate)
        )

        self.set_secure_cookie(
            AUTH_REQUEST_COOKIE_NAME,
            jsonify(request_token),
            expires_days=1,
            path=url_path_join(self.base_url, 'hub', 'oauth_callback'),
            httponly=True,
        )
        self.log.info('oauth redirect: %r', redirect)

        self.redirect(redirect)


class MWCallbackHandler(OAuthCallbackHandler):
    """
    Override OAuthCallbackHandler to take out state parameter handling.

    mwoauth doesn't seem to support it for now!
    """

    def check_arguments(self):
        pass

    def get_state_url(self):
        return None


class MWOAuthenticator(OAuthenticator):
    login_handler = MWLoginHandler
    callback_handler = MWCallbackHandler

    user_auth_state_key = "MEDIAWIKI_USER_IDENTITY"

    @default("login_service")
    def _login_service_default(self):
        return os.environ.get("LOGIN_SERVICE", "MediaWiki")

    mw_index_url = Unicode(
        os.environ.get('MW_INDEX_URL', 'https://meta.wikimedia.org/w/index.php'),
        config=True,
        help="""
        Full path to index.php of the MW instance to use to log in
        """,
    )

    executor_threads = Integer(
        12,
        config=True,
        help="""
        Number of executor threads.

        MediaWiki OAuth requests happen in this thread,
        so it is mostly waiting for network replies.
        """,
    )

    executor = Any()

    def _executor_default(self):
        return ThreadPoolExecutor(self.executor_threads)

    def normalize_username(self, username):
        """
        Override normalize_username to avoid lowercasing usernames.
        """
        username = username.replace(' ', '_')
        return username

    def build_access_tokens_request_params(self, handler, data=None):
        """
        We're overriding this method because mediawiki needs a Handshaker object
        to send the tokes request. So, we're building the params directly in the
        `get_token_info`.
        """
        return None

    async def get_token_info(self, handler, params):
        # Exchange the OAuth code for an Access Token
        consumer_token = ConsumerToken(
            self.client_id,
            self.client_secret,
        )

        handshaker = Handshaker(self.mw_index_url, consumer_token)
        request_token = dejsonify(handler.get_secure_cookie(AUTH_REQUEST_COOKIE_NAME))
        handler.clear_cookie(AUTH_REQUEST_COOKIE_NAME)
        access_token = await wrap_future(
            self.executor.submit(
                handshaker.complete, request_token, handler.request.query
            )
        )
        return {"access_token": access_token, "consumer_token": consumer_token}

    async def token_to_user(self, token_info):
        handshaker = Handshaker(self.mw_index_url, token_info["consumer_token"])

        return await wrap_future(
            self.executor.submit(handshaker.identify, token_info["access_token"])
        )

    def build_auth_state_dict(self, token_info, user_info):
        # this shouldn't be necessary anymore,
        # but keep for backward-compatibility
        return {
            'ACCESS_TOKEN_KEY': token_info["access_token"].key,
            'ACCESS_TOKEN_SECRET': token_info["access_token"].secret,
            self.user_auth_state_key: user_info,
        }
